# -*- coding:utf-8 -*-
"""
数据预处理模块
@version: 1.0
@author: kevin
@license: Apache Licence
@contact: liujiezhang@bupt.edu.cn
@site:
@software: Atom
@file: data-help.py
@time: 17/7/7 下午15:21
"""
import logging
import re
import time
from math import ceil
from collections import Counter
from multiprocessing.dummy import Pool as ThreadPool
import matplotlib.pyplot as plt

import jieba
import jieba.posseg as pseg
import numpy as np
import word2vec
import xlrd
from sklearn.feature_extraction.text import (
    CountVectorizer, HashingVectorizer, TfidfVectorizer)
from sklearn.feature_selection import SelectKBest, chi2
import os
import itertools
from os.path import join, exists, split
logging.getLogger().setLevel(logging.INFO)


class Feature(object):
    """
    特征抽取抽象类
    支持一下词表示:
        1.TF词频 表示词向量
        2.TF-IDF 表示词向量
        3.CHI 表示词权重向量
        4.Hash map
        5.Word2Vec 累加表示词向量
    """

    def __init__(self, stop_words_file):
        '''
        初始化
        :param stop_words_file: 停用词路径
        :return:
        '''
        # 停用词表
        self.stop_words = self.stopwords(stop_words_file)

    def tf(self, X_data, test_data=None, n_features=30000):
        '''
        tf词频
        :param X_data:训练文档集,必须
        :param test_data:测试文本或文本集,非必须
        :param n_features:特征数 top_n原则
        :return:特征向量
        '''
        # 局部init
        try:
            if self.tf_vectorizer:
                pass
        except:
            self.tf_vectorizer = CountVectorizer(
                max_features=n_features,
                stop_words=self.stop_words,
                decode_error='strict')
            x_data_vec = self.tf_vectorizer.fit_transform(X_data)
            print("vocabulary length is %s:" %
                  len(self.tf_vectorizer.vocabulary_))

        # 只对测试文本处理时
        if test_data is not None:  # 支持处理单条或多条测试文本
            return self.tf_vectorizer.transform(test_data)

        # 无测试文本时,返回训练集向量
        return x_data_vec

    def tfidf(self,
              X_data,
              test_data=None,
              n_features=30000,
              max_df=0.95,
              min_df=2):
        '''
        tfidf 词向量统计
        :param X_data:训练文本集,必须
        :param test_data:测试文本或文本集,非必须
        :param n_features:最大特征数
        :param max_df:
        :param min_df:
        :return:词向量
        '''
        # 局部init
        try:
            if self.tfidf_vectorizer:
                pass
        except:
            self.tfidf_vectorizer = TfidfVectorizer(
                max_features=n_features,
                stop_words=self.stop_words,
                max_df=max_df,
                min_df=min_df)
            x_data_vec = self.tfidf_vectorizer.fit_transform(X_data)
            print("vocabulary length is %s:" %
                  len(self.tfidf_vectorizer.vocabulary_))

        # 只对测试文本处理时
        if test_data is not None:  # 支持处理单条或多条测试文本
            return self.tfidf_vectorizer.transform(test_data)

        # 无测试文本时,返回训练集向量
        return x_data_vec

    def chi(self, X_data, Y_data, test_data=None, n_features=2000):
        '''
        CHI词向量表示
        :param X_data:训练文本集,必须
        :param Y_data:训练标签集,必须
        :param test_data:测试文本集,非必须
        :param test_label:测试标签集,非必须load_data
        :param n_features:特征数
        :return:特征向量
        '''
        try:
            if self.chi_2:
                pass
        except:
            # 先转化为tfidf向量矩阵
            self.chi_tfidf_vectorizer = TfidfVectorizer(
                stop_words=self.stop_words)
            x_data_tfidf_vec = self.chi_tfidf_vectorizer.fit_transform(X_data)
            print("vocabulary length is %s:" %
                  len(self.chi_tfidf_vectorizer.vocabulary_))

            # 卡方
            self.chi_2 = SelectKBest(chi2, k=n_features)
            x_data_vec = self.chi_2.fit_transform(x_data_tfidf_vec, Y_data)

        # 只对测试文本处理时
        if test_data is not None:  # 支持处理单条或多条测试文本
            test_data_tfidf_vec = self.chi_tfidf_vectorizer.transform(
                test_data)
            return self.chi_2.transform(test_data_tfidf_vec)

        # 无测试文本时,返回训练集向量
        return x_data_vec

    def hash(self, X_data, n_features=2**18):
        '''
        hash 特征向量
        :param X_data: u'content','file','list','filename'
        :param n_features: 特征数目 默认2 ** 18
        :return: 特征向量
        '''
        # hash特征
        vectorizer = HashingVectorizer(
            stop_words=self.stop_words,
            non_negative=True,
            n_features=n_features)
        X_data_vec = vectorizer.transform(X_data)

        return X_data_vec

    def stopwords(self, file):
        '''
        加载停用词表
        :param file:停用词表路径
        :return:停用词列表
        '''
        with open(file, 'r') as f:
            stop_words = list(f.read().splitlines())

        return stop_words

    def wordToVec(self, X_data, model_file='../docs/txtAll.bin'):
        '''
        word2vec 词向量表示
        :param X_data:文档集
        :param model_file:bin文件目录
        :return:词向量
        '''
        # 文档词向量
        x_data_vec = []
        # 局部初始化
        try:
            if self.model:
                pass
        except:
            # Word2Vec向量训练集
            self.model = word2vec.load(model_file)
        for doc in X_data:
            words = segmentByStopwords(doc, self.stop_words)
            # 句子向量
            sent_vec = np.zeros([1, 100])
            for word in words:
                sent_vec += wordToVec(word, self.model)
            x_data_vec.append(sent_vec)

        return x_data_vec


def train_word2vec(sentence_matrix,
                   vocabulary_inv,
                   num_features=100,
                   min_word_count=1,
                   context=10):
    """
    Trains, saves, loads Word2Vec model
    Returns initial weights for embedding layer.

    inputs:
    sentence_matrix # int matrix: num_sentences x max_sentence_len
    vocabulary_inv  # dict {str:int}
    num_features    # Word vector dimensionality
    min_word_count  # Minimum word count
    context         # Context window size
    """
    model_dir = os.path.abspath('../docs') + '/model/w2v_matrix'
    model_name = "{:d}features_{:d}minwords_{:d}context".format(
        num_features, min_word_count, context)
    model_name = join(model_dir, model_name)
    if exists(model_name):
        embedding_model = word2vec.Word2Vec.load(model_name)
        print('Load existing Word2Vec model \'%s\'' % split(model_name)[-1])
    else:
        # Set values for various parameters
        num_workers = 2  # Number of threads to run in parallel
        downsampling = 1e-3  # Downsample setting for frequent words

        # Initialize and train the model
        print('Training Word2Vec model...')
        sentences = [[vocabulary_inv[w] for w in s] for s in sentence_matrix]
        embedding_model = word2vec.Word2Vec(
            sentences,
            workers=num_workers,
            size=num_features,
            min_count=min_word_count,
            window=context,
            sample=downsampling)

        # If we don't plan to train the model any further, calling
        # init_sims will make the model much more memory-efficient.
        embedding_model.init_sims(replace=True)

        # Saving the model for later use. You can load it later using
        # Word2Vec.load()
        if not exists(model_dir):
            os.mkdir(model_dir)
        print('Saving Word2Vec model \'%s\'' % split(model_name)[-1])
        embedding_model.save(model_name)

    # add unknown words
    embedding_weights = [
        np.array([
            embedding_model[w] if w in embedding_model else np.random.uniform(
                -0.25, 0.25, embedding_model.vector_size)
            for w in vocabulary_inv
        ])
    ]
    return embedding_weights


def load_data_and_labels(file_path,
                         split_tag='\t',
                         lbl_text_index=[0, 1],
                         use_jieba_segment=False,
                         is_shuffle=False):
    """
    Loads MR polarity data from files, splits the data into words and generates labels.
    Returns split sentences and labels.
    """

    # Load data from files
    raw_data = [x.decode('utf-8') for x in open(file_path, 'r').readlines()]
    # parse label
    labels = [
        data.strip('\n').split(split_tag)[lbl_text_index[0]]
        for data in raw_data
    ]
    # parse text
    texts = [
        data.strip('\n').split(split_tag)[lbl_text_index[1]]
        for data in raw_data
    ]
    if use_jieba_segment:
        texts = segDataset(texts)

    # Split by words
    # texts = [clean_str(sent) for sent in texts]
    texts = [filter(lambda a: a != '', s.split(" ")) for s in texts]
    # support multi-label
    labels = [filter(lambda a: a != '', s.split(" ")) for s in labels]
    if is_shuffle:
        ind = np.arange(len(texts))
        np.random.shuffle(ind)
        texts = list(np.array(texts)[ind])
        labels = list(np.array(labels)[ind])

    return texts, labels


def load_trn_tst_data_labels(trn_file,
                             tst_file=None,
                             ratio=0.2,
                             split_tag='\t',
                             lbl_text_index=[0, 1],
                             is_shuffle=False):
    """
    Loads train data and test data,return segment words and labels
    if tst_file is None , split train data by ratio
    """
    # train data
    trn_data, trn_labels = load_data_and_labels(
        trn_file, split_tag, lbl_text_index, is_shuffle=is_shuffle)

    # test data
    if tst_file:
        tst_data, tst_labels = load_data_and_labels(
            tst_file, split_tag, lbl_text_index, is_shuffle=is_shuffle)
    else:
        index = np.arange(len(trn_labels))
        np.random.shuffle(index)
        split_n = int(ratio * len(trn_labels))

        trn_data = np.array(trn_data)
        trn_labels = np.array(trn_labels)

        tst_data, tst_labels = trn_data[index[:split_n]], trn_labels[
            index[:split_n]]
        trn_data, trn_labels = trn_data[index[split_n:]], trn_labels[index[
            split_n:]]

    return list(trn_data), list(trn_labels), list(tst_data), list(tst_labels)


def pad_sentences(sentences, padding_word="<PAD/>", mode='max'):
    """
    Pads all sentences to the same length. The length is defined by the longest sentence.
    Returns padded sentences.
    """
    if isinstance(mode, int):
        sequence_length = mode
    elif mode == 'max':
        sequence_length = max(len(x) for x in sentences)
    else:
        sequence_length = sum(len(x) for x in sentences) / len(sentences)
    padded_sentences = []
    for i in range(len(sentences)):
        sentence = sentences[i]
        if len(sentence) < sequence_length:
            num_padding = sequence_length - len(sentence)
            new_sentence = sentence + [padding_word] * num_padding
        else:
            new_sentence = sentence[:sequence_length]
        padded_sentences.append(new_sentence)
    return padded_sentences


def build_vocab(sentences, padding_word="<PAD/>"):
    """
    Builds a vocabulary mapping from word to index based on the sentences.
    Returns vocabulary mapping and inverse vocabulary mapping.
    """
    # Build vocabulary
    word_counts = Counter(itertools.chain(*sentences))
    # Mapping from index to word
    vocabulary_inv = [x[0] for x in word_counts.most_common()]
    # set padding_word id == 0
    vocabulary_inv.remove(padding_word)
    vocabulary_inv.insert(0, padding_word)
    # Mapping from word to index
    vocabulary = {x: i for i, x in enumerate(vocabulary_inv)}
    return [vocabulary, vocabulary_inv]


def build_input_data(sentences, labels, vocabulary, padding_word="<PAD/>"):
    """
    Maps sentencs and labels to vectors based on a vocabulary.
    """
    x = np.array([[
        vocabulary[word] if word in vocabulary else vocabulary[padding_word]
        for word in sentence
    ] for sentence in sentences])
    y = np.array(labels)
    return [x, y]


def load_data(trn_file,
              tst_file=None,
              ratio=0.2,
              split_tag='\t',
              lbl_text_index=[0, 1],
              vocabulary=None,
              vocabulary_inv=None,
              padding_mod='max',
              is_shuffle=True,
              use_jieba_segment=False,
              use_tst=False):
    """
    Loads and preprocessed data for the MR dataset.
    Returns input vectors, labels, vocabulary, and inverse vocabulary.
    """
    # Load and preprocess data
    print("%s  loading train data and label....." %
          time.asctime(time.localtime(time.time())))
    trn_text, trn_labels = load_data_and_labels(
        trn_file,
        split_tag,
        lbl_text_index,
        is_shuffle=is_shuffle,
        use_jieba_segment=use_jieba_segment)
    if tst_file:
        print("%s  loading train data and label....." %
              time.asctime(time.localtime(time.time())))
        tst_text, tst_labels = load_data_and_labels(
            tst_file, split_tag, lbl_text_index, is_shuffle=is_shuffle)
        sentences, labels = trn_text + tst_text, trn_labels + tst_labels
    else:
        sentences, labels = trn_text, trn_labels
    print("%s  padding sentences....." %
          time.asctime(time.localtime(time.time())))
    sentences_padded = pad_sentences(sentences, mode=padding_mod)

    if vocabulary is None or vocabulary_inv is None:
        print("%s  building vocab....." %
              time.asctime(time.localtime(time.time())))
        vocabulary, vocabulary_inv = build_vocab(sentences_padded)

    x, y = build_input_data(sentences_padded, labels, vocabulary)

    if tst_file is None and not use_tst:
        return [x, y, vocabulary, vocabulary_inv]
    elif tst_file:
        split_n = len(trn_text)
    elif use_tst:
        split_n = int(ratio * len(trn_text))

    return x[split_n:], y[
        split_n:], x[:split_n], y[:split_n], vocabulary, vocabulary_inv


def readExcelByCol(file_name, col):
    '''
    按列读取excel数据
    :param file_name: excel路径
    :param col: 列数,可以多列,或者单列
    :return: 迭代器 可调用next(gener)来返回数据
    '''
    wb = xlrd.open_workbook(file_name)
    sheet = wb.sheet_by_index(0)

    if isinstance(col, list):
        for i in col:
            yield sheet.col_values(i)
    elif isinstance(col, int):
        data = sheet.col_values(col)
        yield data


def corpus_balance(datas, labels, mod='max'):
    """对语料进行重采样 或 欠采样."""
    shuffle_indices = np.random.permutation(np.arange(len(datas)))
    shuffle_datas = np.array(datas)[shuffle_indices]
    shuffle_labels = np.array(labels)[shuffle_indices]
    keys = set(labels)
    label_count = Counter(labels)
    # 以语料 最多 为标准，对其他类别进行 过采样
    if mod == 'max':
        max_label = label_count.most_common(1)[0][0]
        keys.remove(max_label)
        for key in keys:
            data = shuffle_datas[shuffle_labels == key].tolist()
            label = shuffle_labels[shuffle_labels == key].tolist()
            ratio = int((label_count[max_label]) * 1. / label_count[key] - 1.5)
            if ratio < 0:
                continue
            for _ in range(ratio):
                datas.extend(data)
                labels.extend(label)
    # 以语料 最少 为标准，对其他类别进行 欠采样
    elif mod == 'min':
        min_label = label_count.most_common()[-1][0]
        datas = shuffle_datas[shuffle_labels == min_label].tolist()
        labels = shuffle_labels[shuffle_labels == min_label].tolist()
        keys.remove(min_label)
        for key in keys:
            data = shuffle_datas[shuffle_labels == key][:label_count[
                min_label]].tolist()
            label = shuffle_labels[shuffle_labels == key][:label_count[
                min_label]].tolist()
            datas.extend(data)
            labels.extend(label)
    # 均衡采样， 对最多的进行欠抽样，对最少的进行重采样
    elif mod == 'average':
        ind = int((len(keys) - 1) / 2.)
        average_label = label_count.most_common()[ind][0]
        datas = shuffle_datas[shuffle_labels == average_label].tolist()
        labels = shuffle_labels[shuffle_labels == average_label].tolist()
        keys.remove(average_label)
        for key in keys:
            ratio = label_count[key] * 1. / label_count[average_label]
            data = shuffle_datas[shuffle_labels == key]
            label = shuffle_labels[shuffle_labels == key]
            if ratio < 0.8:
                data_shape = (
                    int(1. / ratio + 0.2),
                    1) if len(data.shape) == 2 else int(1. / ratio + 0.2)
                label_shape = (
                    int(1. / ratio + 0.2),
                    1) if len(label.shape) == 2 else int(1. / ratio + 0.2)
                data = np.tile(data, data_shape)
                label = np.tile(label, label_shape)
            elif ratio > 1.2:
                ind = int((1. / ratio + 0.2) * len(data))
                data = data[:ind]
                label = label[:ind]
            datas += data.tolist()
            labels += label.tolist()

    return datas, labels


def segment(content, nomial=False):
    '''
    分词
    :param content: 待分词文本
    :param nomial:是否仅仅保留动名词
    :return:'a b c'
    '''
    content = str(content)
    if nomial:
        nomial_words = []
        words = pseg.cut(content)
        for word in words:
            # print(word.word,word.flag)
            if contain(['n', 'v', 'd', 'a'], word.flag):
                nomial_words.append(word.word)
        return ' '.join(nomial_words)
    else:
        words = jieba.lcut(content.strip('\n'), HMM=True)
        return ' '.join(words)


def segDataset(data_set, parrel=False, nomial=False):
    '''
    文档集分词
    :param data_set:文档集
    :param parrel:是否并行分词,windows不支持
    :param nomial:是否仅仅保留动名词
    :return:['a b c','e d g',....]
    '''
    data_cut = []
    start = time.time()
    print('start cut dataset....')
    if parrel:
        p = ThreadPool(4)
        data_cut = p.map(segment, data_set)
        p.close()
        p.join()
    else:
        for content in data_set:
            data_cut.append(segment(content, nomial))
    end = time.time()
    print('cost time %0.2f seconds.' % (end - start))
    return data_cut


def segmentByStopwords(sentence, stop_words):
    '''
    :param sentence:
    :return:
    '''
    words = jieba.lcut(sentence, cut_all=False)
    for _ in range(len(words)):
        word = words.pop(0)
        if word not in stop_words:
            words.append(word)
    return words


def loadStopWords(file_name):
    '''
    加载停用词
    :param file_name: 停用词表路径
    :return: set()
    '''
    stop_words = set()
    with open(file_name, 'rb') as f:
        for line in f:
            word = line.decode('utf-8').strip()
            ['', stop_words.add(word)][word != '']
    return stop_words


def clean_str(s):
    '''
    字符串清洗
    :param s:原始字符串
    :return:
    '''

    pattern = r"[~（）&:,!！﹗?'`%？#.、。，+*/“”•－：;；·—．-]"
    s = re.sub(r"#.*#|＃.*＃", "", str(s))
    s = re.sub(r"【.*?】", "", s)
    s = re.sub(r"@.*?@", "", s)
    s = re.sub(r"[0-9]+", "Number", s)
    s = re.sub(r"\(\)|（）|__*", "Blank", s)
    s = re.sub(pattern, "", s)
    s = re.sub("nbsp;|amp;|quot;|qpos;", "", s)
    s = re.sub(r"\u2022", "", s)
    s = re.sub(r"大同区同福社区|青年之声", "", s)

    return s.strip()


def save_var(file_path, var):
    with open(file_path, 'w') as f:
        var = ['%s\n' % x.encode('utf-8') for x in var]
        f.writelines(var)


def load_var(file_path):
    with open(file_path, 'r') as f:
        var = [x.decode('utf8').strip('\n') for x in f.readlines()]
        return var


def is_uchar(uchar):
    """判断一个unicode是否是汉字"""
    return True if uchar >= u'\u4e00' and uchar <= u'\u9fa5' else False


def contain(_list, _string):
    '''
    检测字符串是否包含某种字符
    :param _list: 字符模式
    :param _string: 匹配字符串
    :return: T or F
    '''
    for _str in _list:
        if _string.find(_str) != -1:
            return True
    return False


def wordToVec(word, model):
    '''
    获取词向量
    :param word: utf-8格式 词
    :return: 词向量一维矩阵
            [0.1,0.3,....]
    '''
    try:
        return model[word]
    except KeyError:
        print('no key of {0}'.format(word))
        return np.zeros([1, 100])


def loadSentimentVector(file_name):
    """
    Load sentiment vector
    [Surprise, Sorrow, Love, Joy, Hate, Expect, Anxiety, Anger]
    """
    contents = [
        line.strip('\n').split() for line in open(file_name, 'r').readlines()
    ]
    sentiment_dict = {
        line[0].decode('utf-8'): [float(w) for w in line[1:]]
        for line in contents
    }
    return sentiment_dict


def plot_confusion_matrix(cm, classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    print(cm)

    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, format(cm[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


class cnn_data_helper(object):
    """
    CNN_LSTM 专属的数据清洗流程
    流程包括如下:
            1.labels对角向量化
            2.构造labels映射字典
            3.原始文本padding长度fixing处理
            4.文本向量化
    """

    def __init__(self, word2vec_model_dir='../docs/sentiment/extend_dict.txt'):
        '''
        初始化

        :param word2vec_model_dir: word2vec训练的bin文件路径
        :return:
        '''

        self.word2vec_model = loadSentimentVector(word2vec_model_dir)
        self.vocabulary_inv = self.word2vec_model.keys()
        self.vocabulary_inv = ["<PAD/>"] + self.vocabulary_inv
        self.vocabulary = {
            word: ind
            for ind, word in enumerate(self.vocabulary_inv)
        }

    def fit(self, train_x_raw, train_y_raw, padding_mod='max'):
        '''
        训练集空间向量化

        :param train_x_raw:训练文档-词集
        :param train_y_raw:训练文档标签集
        :return:
        '''

        # labels集合
        self.labels = list(set(train_y_raw))
        num_labels = len(self.labels)
        # 转化为one_hot向量
        one_hot = np.zeros((num_labels, num_labels), int)
        np.fill_diagonal(one_hot, 1)
        # 保存label对应one_hot向量
        self.label_dict = dict(zip(self.labels, one_hot))

        # self.build_vocab(train_x_raw)
        # x 是文档-词向量矩阵 固定维度
        train_x_raw = [[
            self.vocabulary[word] for word in doc_words
            if word in self.vocabulary
        ] for doc_words in train_x_raw]
        train_x = self.pad_sentences(
            train_x_raw, padding_word=0, forced_sequence_length=padding_mod)
        y_raw = [self.label_dict[label] for label in train_y_raw]

        train_x = np.array(train_x)
        train_y = np.array(y_raw)

        return train_x, train_y

    def load_embeddings(self, dim=100):
        '''
        加载词典embedding向量

        :return:
        '''
        word_embeddings = {}
        for word in self.vocabulary:
            try:
                word_embeddings[word] = self.word2vec_model[word]
            except:
                word_embeddings[word] = np.random.uniform(-0.1, 0.1, dim)
        return word_embeddings

    def pad_sentences(self,
                      docs_words,
                      padding_word="<PAD/>",
                      forced_sequence_length='average'):
        '''
        归整文档词序列长度

        :param docs_words:文档-词集
        :param padding_word: 默认添加的词
        :param forced_sequence_length: 训练时制定的词序列长度, int or str 'max' and 'average'
        :return: pad 后的文档-词集
        '''
        """Pad setences during training or prediction"""
        if forced_sequence_length == 'average':
            sequence_length = [len(doc_word) for doc_word in docs_words]
            self.sequence_length = int(
                1.3 * sum(sequence_length) / len(sequence_length))

        elif forced_sequence_length == 'max':
            self.sequence_length = max(
                len(doc_word) for doc_word in docs_words)
        else:
            self.sequence_length = int(forced_sequence_length)

        # 存放 pad后的docs
        padded_sentences = []
        for doc_index in range(len(docs_words)):
            doc_words = docs_words[doc_index]
            num_padding = self.sequence_length - len(doc_words)

            if num_padding < 0:  # Prediction: cut off the sentence if it is longer than the sequence length
                # logging.info(
                #     'This sentence has to be cut off because it is longer than trained sequence length')
                padded_sentence = doc_words[0:self.sequence_length]
            else:
                # 训练语句长度不够的话,补全
                padded_sentence = doc_words + [padding_word] * num_padding
            padded_sentences.append(padded_sentence)
        return padded_sentences

    def build_vocab(self, train_x_raw_padded):
        '''
        构建总词典集词典词序映射表

        :param train_x_raw_padded: 文档-fixed词集
        :return:
        '''

        word_counts = Counter(itertools.chain(*train_x_raw_padded))
        self.vocabulary_inv = [word[0] for word in word_counts.most_common()]
        self.vocabulary = {
            word: index
            for index, word in enumerate(self.vocabulary_inv)
        }

    def batch_iter(self, data, batch_size, num_epochs, shuffle=True):
        '''
        数据batch随机化

        :param data: list(zip(x_train, y_train))
        :param batch_size:
        :param num_epochs:
        :param shuffle:
        :return:
        '''
        data = np.array(data)
        data_size = len(data)
        num_batches_per_epoch = int(ceil(float(data_size) / batch_size))

        for epoch in range(num_epochs):
            if shuffle:
                shuffle_indices = np.random.permutation(np.arange(data_size))
                shuffled_data = data[shuffle_indices]
            else:
                shuffled_data = data

            for batch_num in range(num_batches_per_epoch):
                start_index = batch_num * batch_size
                end_index = min((batch_num + 1) * batch_size, data_size)
                yield shuffled_data[start_index:end_index]

    def transforms(self,
                   test_x_raw,
                   padding_word="<PAD/>",
                   forced_sequence_length=None):
        '''
        测试文本向量化
        :param test_x_raw: 测试文本 str or list
        :return:
        '''
        if not isinstance(test_x_raw[0], list):
            test_x_raw = [test_x_raw]

        test_x_raw = self.pad_sentences(
            test_x_raw,
            padding_word=padding_word,
            forced_sequence_length=forced_sequence_length)
        # x 是文档-词向量矩阵 固定维度
        train_x = np.array([[self.vocabulary[word] for word in doc_words]
                            for doc_words in test_x_raw])

        return train_x

    @property
    def _vocabulary(self):
        return self.vocabulary

    @property
    def _vocabulary_inv(self):
        return self.vocabulary_inv

    @property
    def _labels(self):
        return self.labels

    @property
    def _sequence_length(self):
        return self.sequence_length


if __name__ == '__main__':
    pass



# coding: utf-8


def loadSentimentVector(file_name):
    """
    Load sentiment vector
    [Surprise, Sorrow, Love, Joy, Hate, Expect, Anxiety, Anger]
    """
    contents = [
        line.strip('\n').split() for line in open(file_name, 'r').readlines()
    ]
    sentiment_dict = {
        line[0].decode('utf-8'): [float(w) for w in line[1:]]
        for line in contents
    }
    return sentiment_dict


if __name__ == '__main__':
    data_dir = '/Users/zhangliujie/Documents/data/all_vocab'
    # load sentiment_dict
    sentiment_dict = loadSentimentVector('../docs/sentiment/extend_dict.txt')
    cnt = 0
    with open(data_dir+'/40.vocab', 'r') as f:
        for line in f:
            words = line.decode('utf8').split()
            filter_wds = filter(lambda x: x in sentiment_dict, words)

            print('*'*60)
            print(line)
            print(' '.join(filter_wds))

            cnt += 1
            if cnt == 100:
                exit()



"""
Utility functions for constructing MLC models.
"""
from keras.layers import Conv1D, Embedding, Flatten, MaxPool1D
from keras.layers import Dense, Dropout, Input, Activation, Permute, Reshape, Lambda, RepeatVector, merge
from keras.layers import ActivityRegularization, Bidirectional, TimeDistributed
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l2
from keras.layers import GRU
from keras.models import Model
import keras.backend as K


def attention_3d_block(inputs, time_steps, single_attention_vector=False):
    # inputs.shape = (batch_size, time_steps, input_dim)
    print(inputs.shape)
    input_dim = int(inputs.shape[2])
    # time_steps = int(inputs.shape[1])
    a = Permute((2, 1))(inputs)
    a = Reshape((input_dim, time_steps))(a)
    a = Dense(time_steps, activation='softmax')(a)
    if single_attention_vector:
        a = Lambda(lambda x: K.mean(x, axis=1))(a)
        a = RepeatVector(input_dim)(a)
    a_probs = Permute((2, 1))(a)
    output_attention_mul = merge([inputs, a_probs], mode='mul')
    return output_attention_mul


def assemble(name, params):
    if name == 'deep_conv':
        return assemble_deep_conv(params)
    elif name == 'deep_lstm':
        return assemble_deep_lstm(params)
    else:
        raise ValueError("Unknown name of the model: %s." % name)


def assemble_deep_conv(params):
    """
    Construct one of the ADIOS models. The general structure is the following:
                                X-H-(Y0|H0)-H1-Y1,
    where all the H-layers are optional and depend on whether they are
    specified in the params dictionary.
    """
    # X

    input_shape = (params['X']['sequence_length'],
                   params['X']['embedding_dim']
                   ) if params['iter']['model_type'] == "CNN-static" else (
                       params['X']['sequence_length'], )
    X = Input(shape=input_shape, dtype='float32', name='X')

    # embedding
    # Static model do not have embedding layer
    if params['iter']['model_type'] == "CNN-static":
        embedding = X
    elif 'embedding_dim' in params['X'] and params['X']['embedding_dim']:
        embedding = Embedding(
            output_dim=params['X']['embedding_dim'],
            input_dim=params['X']['vocab_size'],
            input_length=params['X']['sequence_length'],
            name="embedding",
            mask_zero=False)(X)
        embedding = Dropout(0.1)(embedding)
    else:
        exit('embedding_dim param is not given!')

    # multi-layer Conv and max-pooling
    conv_layer_num = len(params['Conv1D'])
    for i in range(1, conv_layer_num + 1):
        H_input = embedding if i == 1 else H
        conv = Conv1D(
            filters=params['Conv1D']['layer%s' % i]['filters'],
            kernel_size=params['Conv1D']['layer%s' % i]['filter_size'],
            padding=params['Conv1D']['layer%s' % i]['padding_mode'],
            # activation='relu',
            strides=1,
            bias_regularizer=l2(0.01))(H_input)
        if 'batch_norm' in params['Conv1D']['layer%s' % i]:
            conv = BatchNormalization(
                **params['Conv1D']['layer%s' % i]['batch_norm'])(conv)
        conv = Activation('relu')(conv)
        conv_pooling = MaxPool1D(
            pool_size=params['Conv1D']['layer%s' % i]['pooling_size'],
            strides=1)(conv)
        # dropout
        if 'dropout' in params['Conv1D']['layer%s' % i]:
            H = Dropout(
                params['Conv1D']['layer%s' % i]['dropout'])(conv_pooling)

        # flatten
    H = Flatten(name='H')(H)

    # Y output
    kwargs = params['Y']['kwargs'] if 'kwargs' in params['Y'] else {}
    if 'W_regularizer' in kwargs:
        kwargs['W_regularizer'] = l2(kwargs['W_regularizer'])
    Y = Dense(
        params['Y']['dim'],
        # activation='softmax',
        name='Y_active',
        bias_regularizer=l2(0.01),
        **kwargs)(H)
    # batch_norm
    if 'batch_norm' in params['Y']:
        Y = BatchNormalization(**params['Y']['batch_norm'])(Y)
    Y = Activation('softmax')(Y)
    if 'activity_reg' in params['Y']:
        Y = ActivityRegularization(name='Y', **params['Y']['activity_reg'])(Y)

    return Model(inputs=X, outputs=Y)


def assemble_deep_lstm(params):
    """
    Construct one of the ADIOS models. The general structure is the following:
                                X-H-H1-Y,
    where all the H-layers are optional and depend on whether they are
    specified in the params dictionary.
    """
    # X

    input_shape = (params['X']['sequence_length'],
                   params['X']['embedding_dim'],
                   ) if params['iter']['model_type'] == "CNN-static" else (
                       params['X']['sequence_length'], )
    X = Input(shape=input_shape, dtype='int32', name='X')

    # embedding
    # Static model do not have embedding layer
    if params['iter']['model_type'] == "CNN-static":
        embedding = X
    elif 'embedding_dim' in params['X'] and params['X']['embedding_dim']:
        embedding = Embedding(
            output_dim=params['X']['embedding_dim'],
            input_dim=params['X']['vocab_size'],
            input_length=params['X']['sequence_length'],
            name="embedding",
            mask_zero=False)(X)
        embedding = Dropout(0.1)(embedding)
    else:
        exit('embedding_dim param is not given!')
    # multi-layer LSTM
    lstm_layer_num = len(params['LSTM'])
    # for i in range(1, lstm_layer_num):
    #     lstm_input = embedding if i == 1 else lstm_out
    #     # lstm_input = attention_3d_block(lstm_input, params['X']['sequence_length'], single_attention_vector=False)
    #     lstm_out = Bidirectional(GRU(
    #         params['LSTM']['layer%s' % i]['cell'],
    #         # recurrent_activation='sigmoid',
    #         kernel_regularizer=l2(0.01),
    #         dropout=0.2,
    #         recurrent_dropout=0.2,
    #         return_sequences=True), merge_mode='ave')(lstm_input)
    #
    #     # batch_norm
    #     if 'batch_norm' in params['LSTM']['layer%s' % i]:
    #         kwargs = params['LSTM']['layer%s' % i]['batch_norm']
    #         lstm_out = BatchNormalization(**kwargs)(lstm_out)
    #     # dropout
    #     if 'dropout' in params['LSTM']['layer%s' % i]:
    #         lstm_out = Dropout(
    #             params['LSTM']['layer%s' % i]['dropout'])(lstm_out)
    # last lstm
    # print(lstm_out.shape)
    # exit()
    lstm_out = Bidirectional(GRU(
        params['LSTM']['layer%s' % lstm_layer_num]['cell'],
        # recurrent_activation='sigmoid',
        kernel_regularizer=l2(0.01),
        # kernel_initializer='lecun_normal',
        # recurrent_initializer='glorot_uniform',
        dropout=0.2,
        recurrent_dropout=0.2,
        return_sequences=True
        ), merge_mode='ave')(embedding)
    # attention_mul
    attention_mul = attention_3d_block(lstm_out, params['X']['sequence_length'])
    lstm_out = Flatten()(attention_mul)

    # batch_norm
    # if 'batch_norm' in params['LSTM']['layer%s' % lstm_layer_num]:
    #     kwargs = params['LSTM']['layer%s' % lstm_layer_num]['batch_norm']
    #     lstm_out = BatchNormalization(**kwargs)(lstm_out)
    # # dropout
    # if 'dropout' in params['LSTM']['layer%s' % lstm_layer_num]:
    #     lstm_out = Dropout(
    #         params['LSTM']['layer%s' % lstm_layer_num]['dropout'],
    #         name='H')(lstm_out)

    # ATTENTION PART STARTS HERE
    # attention_probs = Dense(
    #     params['LSTM']['layer%s' % lstm_layer_num]['cell'],
    #     activation='softmax',
    #     name='attention_vec')(lstm_out)
    # lstm_out = merge(
    #     [lstm_out, attention_probs],
    #     output_shape=8,
    #     name='attention_mul',
    #     mode='mul')
    # ATTENTION PART FINISHES HERE
    # Y output
    kwargs = params['Y']['kwargs'] if 'kwargs' in params['Y'] else {}
    if 'W_regularizer' in kwargs:
        kwargs['W_regularizer'] = l2(kwargs['W_regularizer'])
    Y = Dense(
        params['Y']['dim'],
        # activation='sigmoid',
        name='Y_active',
        bias_regularizer=l2(0.01),
        **kwargs)(lstm_out)
    # batch_norm
    if 'batch_norm' in params['Y']:
        Y = BatchNormalization(**params['Y']['batch_norm'])(Y)
    Y = Activation('softmax')(Y)
    if 'activity_reg' in params['Y']:
        Y = ActivityRegularization(name='Y', **params['Y']['activity_reg'])(Y)

    return Model(inputs=X, outputs=Y)



# -*- coding:utf-8 -*-
"""
@version: 1.0
@author: kevin
@license: Apache Licence
@contact: liujiezhang@bupt.edu.cn
@site:
@software: PyCharm Community Edition
@file: adios_train.py
@time: 17/05/03 17:39
"""

import os
import sys

import yaml
import numpy as np
import json
import pickle
from copy import deepcopy
import matplotlib.pyplot as plt
from collections import Counter
from sklearn import preprocessing
from sklearn.metrics import classification_report, confusion_matrix

from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras.optimizers import Adagrad, RMSprop

sys.path.append('..')
from utils.assemble import assemble
from utils.dataHelper import load_data, save_var, corpus_balance, pad_sentences, plot_confusion_matrix


def train(train_dataset,
          valid_dataset,
          test_dataset,
          params,
          model_flag='deep_lstm'):

    # Assemble and compile the model
    model = assemble(model_flag, params)
    raw_test_dataset = deepcopy(test_dataset)
    # Prepare embedding layer weights and convert inputs for static model
    model_type = params['iter']['model_type']
    print("Model type is", model_type)
    if model_type == "CNN-non-static" or model_type == "CNN-static":
        embedding_weights = [
            np.array([
                np.array(sentiment_dict[w]) if w in sentiment_dict else
                np.random.uniform(0, 0., params['X']['embedding_dim']) for w in vocabulary_inv
            ])
        ]

        if model_type == "CNN-static":
            train_dataset['X'] = embedding_weights[0][train_dataset['X']]
            test_dataset['X'] = embedding_weights[0][test_dataset['X']]
            valid_dataset['X'] = embedding_weights[0][valid_dataset['X']]
        elif params['iter']['model_type'] == "CNN-non-static":
            embedding_layer = model.get_layer('embedding')
            embedding_layer.set_weights(embedding_weights)
    elif model_type == "CNN-rand":
        embedding_weights = None
    else:
        raise ValueError("Unknown model type")
    # complie model
    optimizer = RMSprop if 'lstm' in model_flag else Adagrad
    pickle.dump(params, open('../docs/model/params', 'wb'))
    model.compile(
        loss=params['Y']['loss_func'],
        metrics=['accuracy'],
        optimizer=optimizer(params['iter']['learn_rate']))

    # Make sure checkpoints folder exists
    model_dir = params['iter']['model_dir']
    model_name = params['iter']['model_name']
    if not os.path.isdir(model_dir):
        os.makedirs(model_dir)

    # Setup callbacks
    callbacks = [
        ModelCheckpoint(
            model_dir + model_name,
            monitor='val_loss',
            verbose=0,
            save_best_only=True,
            mode='min'),
        EarlyStopping(monitor='val_loss', patience=25, verbose=1, mode='min'),
    ]

    # Fit the model to the data
    batch_size = params['iter']['batch_size']
    nb_epoch = params['iter']['epoch']

    # start to train
    model.fit(
        x=train_dataset["X"],
        y=train_dataset['Y'],
        validation_data=(valid_dataset["X"], valid_dataset["Y"]),
        batch_size=batch_size,
        epochs=nb_epoch,
        callbacks=callbacks,
        verbose=1)

    # Load the best weights
    if os.path.isfile(model_dir + model_name):
        model.load_weights(model_dir + model_name)
    # Test the model
    probs = model.predict(
        test_dataset, verbose=1, batch_size=params['iter']['batch_size'])
    preds = np.argmax(probs, axis=1).flatten()
    targets = np.argmax(test_dataset['Y'], axis=1).flatten()
    for i in range(60):
        print('\n')
        print(' '.join([
            vocabulary_inv[ii]
            for ii in filter(lambda ind: ind != 0, raw_test_dataset['X'][i])
        ]))
        print('target: %s' % category[targets[i]])
        print('predict: %s' % category[preds[i]])
        print('probs: %s' % probs[i])

    print(classification_report(targets, preds, target_names=category))

    # confusion_matrix
    cnf_matrix = confusion_matrix(targets, preds)
    np.set_printoptions(precision=2)
    plt.figure()
    plot_confusion_matrix(cnf_matrix, classes=category, title='Confusion matrix, without normalization')
    plt.figure()
    plot_confusion_matrix(cnf_matrix, classes=category, normalize=True, title='Confusion matrix, with normalization')
    plt.show()


def loadSentimentVector(file_name):
    """
    Load sentiment vector
    [Surprise, Sorrow, Love, Joy, Hate, Expect, Anxiety, Anger]
    """
    contents = [
        line.strip('\n').split() for line in open(file_name, 'r').readlines()
    ]
    sentiment_dict = {
        line[0].decode('utf-8'): [float(w) for w in line[1:]]
        for line in contents
    }
    return sentiment_dict


if __name__ == '__main__':
    # load sentiment vector
    sentiment_dict = loadSentimentVector('../docs/sentiment/extend_dict.txt')
    # use sentiment_dict as vocabulary
    vocabulary_inv = ["<PAD/>"] + list(sentiment_dict.keys())
    vocabulary = {x: i for i, x in enumerate(vocabulary_inv)}

    # Load the datasets
    texts, labels, vocabulary, vocabulary_inv = load_data(
        '../docs/sentiment/data.vocab',
        use_tst=False,
        lbl_text_index=[0, 1],
        split_tag='@@@',
        padding_mod='max',
        use_jieba_segment=False,
        vocabulary=vocabulary,
        vocabulary_inv=vocabulary_inv
        )

    labels = [x[0] for x in labels]
    # filter by sentiment_dict
    texts = [filter(lambda x: x != 0, content) for content in texts]

    # padding
    texts = pad_sentences(texts, padding_word=0, mode=80)
    print(Counter(labels))
    # texts, labels = corpus_balance(texts, labels, mod='average')
    # save vocab
    save_var('../docs/model/vocabulary_inv', vocabulary_inv)
    category = ['1', '2', '3', '4', '5']
    # [0] -> 0
    print(Counter(labels))
    # save category
    save_var('../docs/model/category', category)
    # vectorize
    labelEncoder = preprocessing.label_binarize
    _labels = labelEncoder(labels, classes=category)
    _texts = np.array(texts)
    # print(_labels[100:300])
    # exit()

    ind = np.arange(_texts.shape[0])
    np.random.shuffle(ind)
    ratio = 0.3
    split_n = int(ratio * len(_texts))
    trn_labels = _labels[ind[split_n:]]
    trn_text = _texts[ind[split_n:]]
    tst_labels = _labels[ind[:split_n]]
    tst_text = _texts[ind[:split_n]]

    print('train data size : %d , test data size : %d' % (len(trn_labels),
                                                          len(tst_labels)))
    print('X sequence_length is : %d , Y dim : %d' % (trn_text.shape[1],
                                                      trn_labels.shape[1]))
    # load params config
    params = yaml.load(open('../utils/config.yaml'))
    params['X']['sequence_length'] = trn_text.shape[1]
    params['X']['vocab_size'] = len(vocabulary)
    params['Y']['dim'] = len(category)

    print(trn_text[0])
    print(trn_labels[0])
    ratio = 0.5
    valid_N = int(ratio * tst_text.shape[0])
    train_dataset = {'X': trn_text, 'Y': trn_labels}
    valid_dataset = {'X': tst_text[:valid_N], 'Y': tst_labels[:valid_N, :]}
    test_dataset = {'X': tst_text[valid_N:], 'Y': tst_labels[valid_N:, :]}
    # start train
    train(train_dataset, valid_dataset, test_dataset, params, model_flag='deep_lstm')



# -*- coding:utf-8 -*-

import os
import pickle

from . import utils

import jieba
import numpy as np

__version__ = '0.0.9'

MODEL_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'docs/model')


class UgcSentiment(object):
    def __init__(self, model_flag='deep_lstm', model_dir=MODEL_DIR):
        self.params = pickle.load(open(model_dir + '/params', 'rb'))
        self.vocabulary_inv = utils.dataHelper.load_var(
            model_dir + '/vocabulary_inv')
        self.vocabulary = {wd: i for i, wd in enumerate(self.vocabulary_inv)}
        self.labels = [
            x.strip('\n') for x in open(model_dir + '/category', 'r')
        ]
        # assemble model
        self.model = utils.assemble.assemble(model_flag, self.params)
        # load trained weights
        self.model.load_weights(model_dir + '/default_best.h5')

    def sentiment(self, datas):
        """
        1. process data
        2. predict probs
        3. return json data
        """
        process_data = self.process(datas)
        probs = self.model.predict(np.array(process_data))
        # probs = probs * (probs > 0.)
        # probs[:, 1] += 0.6
        # probs[:, 2] -= 0.3
        preds = np.argmax(probs, axis=1).flatten()
        # return result
        result = []
        for i in range(len(probs)):
            json_data = {
                "labels": self.labels[preds[i]],
                "probs_json": {self.labels[j]: probs[i][j]
                               for j in range(5)},
                "probs": ' '.join(probs[i].astype('string'))
            }
            result.append(json_data)
        # single data return json
        if not isinstance(datas, list):
            return result[0]
        # multi data return list of json
        return result

    def process(self, datas):
        """
        1. segment using jieba
        2. transform word to word_index
        3. padding_word
        """
        if not isinstance(datas, list):
            datas = [datas]
        self.seg_data = [jieba.lcut(data) for data in datas]
        data = [[self.vocabulary[wd] for wd in sen if wd in self.vocabulary]
                for sen in self.seg_data]
        pad_data = utils.dataHelper.pad_sentences(
            data, padding_word=0, mode=self.params['X']['sequence_length'])

        return pad_data

    @property
    def words(self):
        return self.seg_data



# coding: utf-8
import os
import sys
import subprocess
import ugc_sentiment
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup
    from distutils.util import convert_path


def _find_packages(where='.', exclude=()):
    """Return a list all Python packages found within directory 'where'
        'where' should be supplied as a "cross-platform" (i.e. URL-style)
        path; it will be converted to the appropriate local path syntax.
        'exclude' is a sequence of package names to exclude; '*' can be used
        as a wildcard in the names, such that 'foo.*' will exclude all
        subpackages of 'foo' (but not 'foo' itself).
        """
    out = []
    stack = [(convert_path(where), '')]
    while stack:
        where, prefix = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if ('.' not in name and os.path.isdir(fn) and
                    os.path.isfile(os.path.join(fn, '__init__.py'))):
                out.append(prefix + name)
                stack.append((fn, prefix + name + '.'))
    for pat in list(exclude) + ['ez_setup', 'distribute_setup']:
        from fnmatch import fnmatchcase
        out = [item for item in out if not fnmatchcase(item, pat)]


PUBLISH_CMD = 'python setup.py register sdist upload'

if 'publish' in sys.argv:
    status = subprocess.call(PUBLISH_CMD, shell=True)
    sys.exit(status)

setup(
    name="ugc-sentiment",
    version=ugc_sentiment.__version__,
    author="Liujie zhang",
    author_email='liujiezhangbupt@gmail.com',
    description="sentiment classifier for ugc comment trained by LSTM with attention machanism.",
    keywords=['sentiment', 'LSTM', 'ugc', 'attention'],
    license="MIT",
    url="https://github.com/KillersDeath/ugc-sentiment",
    packages=find_packages(exclude=('test*', )),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2.7",
    ],
    include_package_data=True
    )
